import math
cdef extern from "math.h":
    double sin(double c)
    double cos(double c)
    double sqrt(double c)

cpdef float distance2(double p1x, double p1y, double p2x, double p2y):
    return sqrt((p2x-p1x)**2 + (p2y-p1y)**2)

cdef class Vector2:
# defined in .pxd
#    cdef public double x, y

    def __cinit__(self, double x, double y):
        self.x = x
        self.y = y

    def __copy__(self):
        return Vector2(self.x, self.y)

    copy = __copy__

    def __repr__(self):
        return 'Vector2(%.2f, %.2f)' % (self.x, self.y)

    '''
    def __eq__(self, other):
        if isinstance(other, Vector2):
            return self.x == other.x and \
                   self.y == other.y
        else:
            assert hasattr(other, '__len__') and len(other) == 2
            return self.x == other[0] and \
                   self.y == other[1]

    def __neq__(self, other):
        return not self.__eq__(other)
    '''

    def __nonzero__(self):
        return self.x != 0 or self.y != 0

    def __len__(self):
        return 2

    def __getitem__(self, key):
        return (self.x, self.y)[key]

    def __setitem__(self, key, value):
        l = [self.x, self.y]
        l[key] = value
        self.x, self.y = l

    def __iter__(self):
        return iter((self.x, self.y))

    def __add__(self, other):
        if isinstance(other, Vector2):
            # Vector + Vector -> Vector
            # Vector + Point -> Point
            # Point + Point -> Vector
            if self.__class__ is other.__class__:
                _class = Vector2
            else:
                from gamelib.euclid import Point2
                _class = Point2
            return _class(self.x + other.x,
                          self.y + other.y)
        else:
            assert hasattr(other, '__len__') and len(other) == 2
            return Vector2(self.x + other[0],
                           self.y + other[1])
    __radd__ = __add__

    def __iadd__(self, other):
        if isinstance(other, Vector2):
            self.x += other.x
            self.y += other.y
        else:
            self.x += other[0]
            self.y += other[1]
        return self

    def __sub__(self, other):
        if isinstance(other, Vector2):
            # Vector - Vector -> Vector
            # Vector - Point -> Point
            # Point - Point -> Vector
            if self.__class__ is other.__class__:
                _class = Vector2
            else:
                from gamelib.euclid import Point2
                _class = Point2
            return _class(self.x - other.x,
                          self.y - other.y)
        else:
            assert hasattr(other, '__len__') and len(other) == 2
            return Vector2(self.x - other[0],
                           self.y - other[1])

   
    def __rsub__(self, other):
        if isinstance(other, Vector2):
            return Vector2(other.x - self.x,
                           other.y - self.y)
        else:
            assert hasattr(other, '__len__') and len(other) == 2
            return Vector2(other.x - self[0],
                           other.y - self[1])

    def __mul__(self, double other):
        return Vector2(self.x * other, self.y * other)

    __rmul__ = __mul__

    def __imul__(self, double other):
        self.x *= other
        self.y *= other
        return self

    def __div__(self, double other):
        return Vector2(self.x / other, self.y / other)


    def __rdiv__(self, double other):
        return Vector2(other / self.x, other / self.y)

    def __floordiv__(self, double other):
        return Vector2(self.x // other, self.y // other)

    def __rfloordiv__(self, double other):
        return Vector2(other // self.x, other // self.y)

    def __truediv__(self, double other):
        return Vector2(self.x / other, self.y / other)

    def __rtruediv__(self, double other):
        return Vector2(other / self.x, other / self.y)
    
    def __neg__(self):
        return Vector2(-self.x, -self.y)

    __pos__ = __copy__
    
    def __abs__(self):
        return sqrt(self.x ** 2 + self.y ** 2)

    magnitude = __abs__

    def magnitude_squared(self):
        return self.x ** 2 + self.y ** 2

    def normalize(self):
        d = sqrt(self.x ** 2 + self.y ** 2)
        if d:
            self.x /= d
            self.y /= d
        return self

    def normalized(self):
        d = sqrt(self.x ** 2 + self.y ** 2)
        if d:
            return Vector2(self.x / d, self.y / d)
        return self.copy()

    def dot(self, Vector2 other):
        return self.x * other.x + self.y * other.y

    def cross(self):
        return Vector2(self.y, -self.x)

    def reflect(self, Vector2 normal):
        # assume normal is normalized
        d = 2 * (self.x * normal.x + self.y * normal.y)
        return Vector2(self.x - d * normal.x, self.y - d * normal.y)

cdef class Matrix3:
# defined in .pxd
#    cdef public double a, b, c, d, e, f, g, h, i, j, k

    def __cinit__(self):
        self.identity()

    def __copy__(self):
        cdef Matrix3 M = Matrix3()
        M.a = self.a
        M.b = self.b
        M.c = self.c
        M.e = self.e 
        M.f = self.f
        M.g = self.g
        M.i = self.i
        M.j = self.j
        M.k = self.k
        return M

    copy = __copy__
    def __repr__(self):
        return ('Matrix3([% 8.2f % 8.2f % 8.2f\n'  \
                '         % 8.2f % 8.2f % 8.2f\n'  \
                '         % 8.2f % 8.2f % 8.2f])') \
                % (self.a, self.b, self.c,
                   self.e, self.f, self.g,
                   self.i, self.j, self.k)

    def __getitem__(self, key):
        return [self.a, self.e, self.i,
                self.b, self.f, self.j,
                self.c, self.g, self.k][key]

    def __setitem__(self, key, value):
        L = self[:]
        L[key] = value
        (self.a, self.e, self.i,
         self.b, self.f, self.j,
         self.c, self.g, self.k) = L

    def __mul__(self, other):
        cdef Vector2 B, V
        cdef double Aa, Ab, Ac, Ae, Af, Ag, Ai, Aj, Ak
        cdef double Ba, Bb, Bc, Be, Bf, Bg, Bi, Bj, Bk
        cdef Matrix3 C

        if isinstance(other, Matrix3):
            # Caching repeatedly accessed attributes in local variables
            # apparently increases performance by 20%.  Attrib: Will McGugan.
            Aa = self.a; Ab = self.b; Ac = self.c; Ae = self.e; Af = self.f; Ag = self.g; Ai = self.i; Aj = self.j; Ak = self.k
            Ba = other.a; Bb = other.b; Bc = other.c; Be = other.e; Bf = other.f; Bg = other.g; Bi = other.i; Bj = other.j; Bk = other.k
            C = Matrix3()
            C.a = Aa * Ba + Ab * Be + Ac * Bi
            C.b = Aa * Bb + Ab * Bf + Ac * Bj
            C.c = Aa * Bc + Ab * Bg + Ac * Bk
            C.e = Ae * Ba + Af * Be + Ag * Bi
            C.f = Ae * Bb + Af * Bf + Ag * Bj
            C.g = Ae * Bc + Af * Bg + Ag * Bk
            C.i = Ai * Ba + Aj * Be + Ak * Bi
            C.j = Ai * Bb + Aj * Bf + Ak * Bj
            C.k = Ai * Bc + Aj * Bg + Ak * Bk
            return C
        if isinstance(other, Vector2):
            A = self
            B = other
            V = Vector2(0, 0)
            V.x = A.a * B.x + A.b * B.y 
            V.y = A.e * B.x + A.f * B.y 
            return V
        from gamelib.euclid import Point2
        if isinstance(other, Point2):
            A = self
            B = other
            P = Point2(0, 0)
            P.x = A.a * B.x + A.b * B.y + A.c
            P.y = A.e * B.x + A.f * B.y + A.g
            return P
        else:
            other = other.copy()
            other._apply_transform(self)
            return other

    def __imul__(self, Matrix3 other):
        cdef double Aa, Ab, Ac, Ae, Af, Ag, Ai, Aj, Ak
        cdef double Ba, Bb, Bc, Be, Bf, Bg, Bi, Bj, Bk
        # Cache attributes in local vars (see Matrix3.__mul__).
        Aa = self.a; Ab = self.b; Ac = self.c; Ae = self.e; Af = self.f
        Ag = self.g; Ai = self.i; Aj = self.j; Ak = self.k; Ba = other.a
        Bb = other.b; Bc = other.c; Be = other.e; Bf = other.f; Bg = other.g
        Bi = other.i; Bj = other.j; Bk = other.k
        self.a = Aa * Ba + Ab * Be + Ac * Bi
        self.b = Aa * Bb + Ab * Bf + Ac * Bj
        self.c = Aa * Bc + Ab * Bg + Ac * Bk
        self.e = Ae * Ba + Af * Be + Ag * Bi
        self.f = Ae * Bb + Af * Bf + Ag * Bj
        self.g = Ae * Bc + Af * Bg + Ag * Bk
        self.i = Ai * Ba + Aj * Be + Ak * Bi
        self.j = Ai * Bb + Aj * Bf + Ak * Bj
        self.k = Ai * Bc + Aj * Bg + Ak * Bk
        return self

    def identity(self):
        self.a = self.f = self.k = 1.
        self.b = self.c = self.e = self.g = self.i = self.j = 0
        return self

    def scale(self, double x, double y):
        # *= not handled well by cython (generates pointer warning)
        return self.__imul__(Matrix3.new_scale(x, y))

    def translate(self, double x, double y):
        return self.__imul__(Matrix3.new_translate(x, y))

    def rotate(self, double angle):
        return self.__imul__(Matrix3.new_rotate(angle))

    # Static constructors
    @classmethod
    def new_identity(cls):
        return Matrix3()

    @classmethod
    def new_scale(cls, double x, double y):
        cdef Matrix3 self = Matrix3()
        self.a = x
        self.f = y
        return self

    @classmethod
    def new_translate(cls, double x, double y):
        cdef Matrix3 self = Matrix3()
        self.c = x
        self.g = y
        return self

    @classmethod
    def new_rotate(cls, double angle):
        cdef Matrix3 self = Matrix3()
        cdef s = sin(angle)
        cdef c = cos(angle)
        self.a = self.f = c
        self.b = -s
        self.e = s
        return self

