import math

from gamelib.euclid import Vector2, Matrix3

def generate(p1x, p1y, p2x, p2y, cp1x, cp1y, cp2x, cp2y, step):
    '''Given the two points p1 and p2 and their control points cp1 and cp2
    generate a cubic bezier curve with steps of "step".

    Return the list of (x, y) points on the curve.
    '''
    l = []
    # generate the cubic bezier points
    x1 = cp1x * 3; x2 = cp2x * 3
    y1 = cp1y * 3; y2 = cp2y * 3
    t = 0
    while t <= (1 + step):
        a = t; a2 = a**2; a3 = a**3
        b = 1 - t; b2 = b**2; b3 = b**3
        px = p1x*b3 + x1*b2*a + x2*b*a2 + p2x*a3
        py = p1y*b3 + y1*b2*a + y2*b*a2 + p2y*a3
        l.append((px, py))
        t += step
    return l

def curve_vertices(x, y, size, start_vector, direction, start_width,
        rotation, num_curves):
    vertices = []
    curves = []

    p1 = Vector2(x, y)
    cp1v = start_vector.normalized()

    rcp = Matrix3.new_rotate(direction * rotation)
    r90 = Matrix3.new_rotate(direction * math.pi/2)
    nr90 = Matrix3.new_rotate(-direction * math.pi/2)

    # p1 -> p2 = d
    d = rcp * start_vector.normalized()
    ds = size
    d = d * ds

    # normal at start of curve
    wn = nr90 * start_vector.normalized()

    # d normalalised for reflection
    dn = d.normalized()

    wstart = start_width
    wend = wstart/1.618

    # generate curves while they're greater than 5 pixels in size
    for i in range(num_curves):
        if ds < 10:
            wend = 1

        cpsize = ds/(1.618*1.5)
        cp1 = p1 + cp1v * cpsize

        # p2
        p2 = p1 + d

        # reflect first control point vector against d to get second
        cp2v = cp1v.reflect(dn)# - 2. * (dn.dot(cp1v)) * dn
        cp2 = p2 + cp2v * cpsize

        # figure the first outer point for the first quad
        v = wn * wstart
        ptx, pty = p1.x+v.x, p1.y+v.y

        # save off the config for this curve
        curves.append((p1, p2, cp1, cp2, Vector2(ptx, pty), d, wstart))

        x1 = cp1.x * 3; x2 = cp2.x * 3
        y1 = cp1.y * 3; y2 = cp2.y * 3
        last = None
        w = wstart

        points = generate(p1.x, p1.y, p2.x, p2.y, cp1.x, cp1.y,
            cp2.x, cp2.y, 0.05)

        N = len(points)
        for i, (px, py) in enumerate(points):
            if last is not None:
                # figure the next quad to add on
                lx, ly, ltx, lty = last
                # vector from last point to this one
                v = Vector2(px-lx, py-ly).normalized()
                # normal to get width
                wn = nr90 * v 
                # gradually scale width
                t = i/float(N)
                w = wstart * (1-t) + wend*t
                # and this is our actual width vector
                v = wn * w
                # add on to the point
                ptx, pty = px+v.x, py+v.y
                # and we have four quad corners
                vertices.extend([lx, ly, ltx, lty, ptx, pty, px, py])

            last = (px, py, ptx, pty)

        # next segment
        p1 = p2
        cp1v = -cp2v
        # the old d is the new d normal away from curve direction
        d = (r90 * d) / 1.618
        dn = d.normalized()
        ds = abs(d)
        wstart, wend = (wend, wend/1.618)

    colors = [0, 0, 0, 0]*N
    return curves, vertices, colors

def path_vertices(x, y, size, start_vector, direction, wstart, wend, type,
        rotation):
    rcp = Matrix3.new_rotate(direction * rotation)
    r90 = Matrix3.new_rotate(direction * math.pi/2)
    nr90 = Matrix3.new_rotate(-direction * math.pi/2)

    cp1v = start_vector

    # p1 -> p2 = d
    start_point = p1 = Vector2(x, y)
    d = rcp * cp1v.normalized()
    ds = size
    d = d * ds

    # d normalalised for reflection
    dn = d.normalized()

    # generate one curve
    cpsize = ds/2
    if type == 1:
        # reflect first control point vector against d to get second so
        # we curve
        cp2v = cp1v.reflect(dn)
    else:
        # reverse first control point so we wriggle
        cp2v = -cp1v
    cp1 = p1 + cp1v * cpsize

    # p2
    p2 = p1 + d
    cp2 = p2 + cp2v * cpsize

    points = generate(p1.x, p1.y, p2.x, p2.y, cp1.x, cp1.y,
        cp2.x, cp2.y, 0.02)
    end_point = Vector2(*points[-1])

    v = start_vector
    v.normalize()
    r90 = Matrix3.new_rotate(direction * math.pi/2)
    n = r90*v * wstart/2
    branch_points = []
    p = start_point
    lines = []
    quads = []
    N = float(len(points))
    for i, (x, y) in enumerate(points):
        if i == 0: continue
        sp1 = p - n
        sp2 = p + n

        t = i/N
        width = wstart * (1-t) + wend*t

        # move along
        p2 = Vector2(x, y)
        v = (p2 - p).normalized()
        n = r90*v * width/2
        ep1 = p2 - n
        ep2 = p2 + n

        quads.extend([
            sp1.x, sp1.y, sp2.x, sp2.y,
            ep2.x, ep2.y, ep1.x, ep1.y
        ])
        lines.extend([
            sp1.x, sp1.y, ep1.x, ep1.y,
            sp2.x, sp2.y, ep2.x, ep2.y
        ])

        p = p2

        # save off the branching point
        if i == 1 and not branch_points:
            branch_points.append((ep1, p2, ep2, v, i))
        elif t > 0.618 and len(branch_points) == 1:
            branch_points.append((ep1, p2, ep2, v, i))

    end_vector = -v

    return points, branch_points, quads, lines, end_vector

try:
    from gamelib import _curve
    generate = _curve.generate
    curve_vertices = _curve.curve_vertices
    path_vertices = _curve.path_vertices
    accelerated = True
except Exception, e:
    print '_curve not available:', e
    accelerated = False


def speed_test():
    generate(0, 0, 5, 0, 0, 1, 5, -1, .0001)


