import math
import random

from euclid import *

import cocos

import pyglet
from pyglet.window import key

HURT_ZAPS = 256

class PlayerLayer(cocos.layer.ScrollableLayer):
    is_event_handler = True

    def __init__(self):
        super(PlayerLayer, self).__init__()

        # initial player position
        self.player = cocos.sprite.Sprite('player-middle.png')
        self.add(self.player, z=1)
        self.player_outer = cocos.sprite.Sprite('player-outer.png',
            anchor=(0, 16))
        self.player_outer.do(cocos.actions.Repeat(cocos.actions.RotateBy(-360, 1)))
        self.player.add(self.player_outer, z=1)
        self.player_inner = cocos.sprite.Sprite('player-inner.png',
            anchor=(14, 12))
        self.player_inner.do(cocos.actions.Repeat(cocos.actions.RotateBy(360, 1)))
        self.player.add(self.player_inner, z=1)

        self.player.color = (255, 255, 255)
        self.player_outer.color = (255, 255, 255)
        self.player_inner.color = (255, 255, 255)

        # player movement
        self.player_velocity = Vector2(50, 0)
        self.path_index = None

        # zappy lines
        self.push_time = 0
        l = [0]*32
        c = [255, 255, 255, 0, 255, 255, 255, 255]*8
        self.lines = self.batch.add(16, pyglet.gl.GL_LINES, None,
            ('v2f', l), ('c4B', c))

        # hurt zaps
        l = [0]*(HURT_ZAPS*2)
        c = [255, 255, 0, 255, 255, 0, 0, 50]*(HURT_ZAPS//2)
        self.hurt_zaps = self.batch.add(HURT_ZAPS, pyglet.gl.GL_LINES, None,
            ('v2f', l), ('c4B', c))
        self.hurt_zaps_used = []
        self.hurt_age = [0]*HURT_ZAPS

    def on_enter(self):
        super(PlayerLayer, self).on_enter()
        self.keys = key.KeyStateHandler()
        cocos.director.director.window.push_handlers(self.keys)
        self.game = self.get_ancestor(cocos.scene.Scene)
        
        if self.path_index is None:
            self.path_index = 0

            # initial path / player position
            x1, y1 = self.game.tree.path[0][-2:]
            self.player.position = x1, y1
            self.path_position = x1, y1

            # player rotation
            #x2, y2 = self.game.tree.path[1][-2:]
            #self.player.rotation = 90-math.degrees(math.atan2(y2-y1, x2-x1))

        self.parent.set_focus(*self.player.position)

    def on_exit(self):
        super(PlayerLayer, self).on_exit()
        self.game = None
        cocos.director.director.window.pop_handlers()

    health = 1
    def hurt(self, amount):
        if self.health == 0:
            self.game.game_over()

        self.health = max(0, self.health - amount)
        
        if self.health > .5:
            b = 100 + max(0, int((self.health-.5) * 2 * 155))
            self.player_outer.opacity = (self.health-.5) * 2
        else:
            self.player_outer.opacity = 0
            b = 100
            self.player_inner.opacity = self.health* 2
        g = 100 + int(self.health * 155)
        r = 255

        self.player.color = (r, g, b)
        self.player_outer.color = (r, g, b)
        self.player_inner.color = (r, g, b)

        # add a hurt zap
        n = list(set(range(HURT_ZAPS)) - set(self.hurt_zaps_used))[0]
        x, y = self.player.position
        x2, y2 = x - 40 + random.random()*80, y-40 +random.random()*80
        self.hurt_zaps.vertices[n*4:n*4+4] = [x2, y2, x, y]
        self.hurt_zaps.colors[n*8+3:(n+1)*8+3:4] = [255, 55]
        self.hurt_zaps_used.append(n)
        self.hurt_age[n] = 1.

    if __debug__:
        def on_key_press(self, k, modifiers):
            if k == key.SPACE:
                self.hurt(.1)

    move_out = 0
    velocity = 100
    current_path = None
    def update(self, dt):
        if self.game.starting:
            return

        # accelerate
        move = (self.keys[key.UP] - self.keys[key.DOWN]) * 200
        if not move:
            move = (self.keys[key.W] - self.keys[key.S]) * 200
        self.velocity = min(150, max(50, self.velocity + move*dt))
        move = self.velocity*dt

        self.game.add_score(move/100)

        index = self.path_index
        position = self.path_position
        self.path_index, self.path_position, segment = \
            self.game.tree.move_along_path(move, position, index, self)

        # left/right displacement
        push = (self.keys[key.RIGHT] - self.keys[key.LEFT]) * 200 * dt
        if not move:
            push = (self.keys[key.D] - self.keys[key.A]) * 200 * dt

        if push:
            self.push_time += dt
        else:
            self.push_time = 0

        if self.move_out < 0:
            self.move_out += push + ((self.move_out ** 2)/5*dt)
        else:
            self.move_out += push - ((self.move_out ** 2)/5*dt)

        # displacement vector == normal to segment
        n = Matrix3.new_rotate(-math.pi/2) * segment
        d = n.normalized() * self.move_out
        lx, ly = self.path_position
        x, y = self.player.position = lx + d.x, ly + d.y

        # draw the zappy lines
        segment.normalize()
        segment *= 20
        p = Vector2(*self.path_position)
        l = []
        for i in range(8):
            t = i/8.
            t -= .5
            p1 = p + segment*t
            l.extend([p1.x, p1.y, x, y])
        self.lines.vertices[:] = l

        # color them
        t = 1. - min(1., self.push_time / 2.)
        if not t:
            self.hurt(.01)
        if t > .5:
            b = 100 + max(0, int((t-.5) * 2 * 155))
        else:
            b = 100
        g = 100 + int(t * 155)
        r = 255
        c = [r, g, b, 0, r, g, b, 255]*8
        self.lines.colors[:] = c

        # age hurt zaps
        for n in list(self.hurt_zaps_used):
            age = self.hurt_age[n]
            self.hurt_age[n] = age = max(0, age-dt)
            if age == 0:
                self.hurt_zaps.vertices[n*4:n*4+4] = [0]*4
                self.hurt_zaps_used.remove(n)
            a = int(255 * age)
            self.hurt_zaps.colors[n*8+3:(n+1)*8+3:4] = [a, max(0,a-200)]

        # scrolling focus
        self.parent.set_focus(*self.path_position)


# baddies could:
# . throb and rotate
# . animate
class Baddie(cocos.sprite.Sprite):
    wobble = False
    wobble_amount = 0
    def update(self, dt, game):
        if self.visible:
            bx, by = self.position
            px, py = game.player.player.position
            d = distance2(px, py, bx, by)
            if d < 25:
                game.player.hurt((20-d) * dt / 5)

        info = game.tree.move_along_path(200*dt, self.path_position, self.path_index)
        if info is None:
            self.parent.remove(self)
            game.unschedule(self.update)
            return
        self.path_index, self.path_position, segment = info
        if self.path_index == 0:
            self.parent.remove(self)
            game.unschedule(self.update)
            return
        #self.status_text.element.text = '%s/%s'%(self.start_index, self.path_index)

        if self.wobble:
            self.wobble_amount += dt*self.wobble_speed
            if self.wobble_amount > math.pi*2:
                self.wobble_amount -= math.pi*2
            n = Matrix3.new_rotate(-math.pi/2) * segment
            d = n.normalized() * (math.sin(self.wobble_amount) * 30)
            x, y = self.path_position
            self.position = x + d.x, y + d.y
        else:
            self.position = self.path_position

        # handle visibility
        path, index, x, y = game.tree.path[self.path_index]
        if path.growth == 1:
            self.visible = True
            self.opacity = path.alphas[index]


class BaddiesLayer(cocos.layer.ScrollableLayer):
    time_to_next = 0
    def update(self, dt):
        if self.time_to_next < 0:
            i = min(len(self.game.tree.path) - 1,
                self.game.player.path_index + 300)
            x, y = self.game.tree.path[i][-2:]
            b = Baddie('baddie.png', (x, y))
            b.start_index = i
            #b.status_text = cocos.text.Label('hello')
            #b.add(b.status_text)
            b.path_position = b.position

            # make things progressively harder
            delay = 4
            if self.game.score > 150:
                b.wobble = True
                b.wobble_speed = 3 + random.random()*2
                delay = 1
            elif self.game.score > 100:
                b.wobble = True
                b.wobble_speed = 3 + random.random()*2
                delay = 2
            elif self.game.score > 50:
                if random.random() > .5:
                    b.wobble = True
                    b.wobble_speed = 2 + random.random()*2
                delay = 3
            elif self.game.score > 20:
                if random.random() > .7:
                    b.wobble = True
                    b.wobble_speed = 2

            self.game.schedule(b.update, self.game)
            b.path_index = i
            b.visible = False
            b.color = (0xbe, 0x36, 0x43)
            self.add(b)

            self.time_to_next = 1 + delay * random.random()
        else:
            self.time_to_next -= dt

    def on_enter(self):
        super(BaddiesLayer, self).on_enter()
        self.game = self.get_ancestor(cocos.scene.Scene)

    def on_exit(self):
        super(BaddiesLayer, self).on_exit()
        self.game = None

